{-|
Module      : OpenAlex
Description : Textmining Collaborative Platform
Copyright   : (c) CNRS, 2023-Present
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX

@Gargantext@: search, map, share
-}

module OpenAlex
  ( module OpenAlex.Client
  , module OpenAlex.Types
  -- , fetchConcepts'
  , fetchConcepts
  , fetchWorks
  , fetchWorksC
  )
where

import Conduit
import Control.Monad.Fail (fail)
import Data.Text qualified as T
import Network.HTTP.Client (newManager, requestHeaders)
import Network.HTTP.Client.TLS (tlsManagerSettings)
import OpenAlex.Client
import OpenAlex.ServantClientLogging
import OpenAlex.Types (ListOf(..), Meta(..), Page, PerPage, Cursor, Filter, Search, Concept, Work, showDate)
import Protolude hiding (yield)
import Servant.Client (BaseUrl(..), ClientEnv(..), ClientError, Scheme(Https), defaultMakeClientRequest, mkClientEnv, runClientM)

defaultClientEnv :: IO ClientEnv
defaultClientEnv = do
  manager <- newManager tlsManagerSettings
  let env' = mkClientEnv manager $ BaseUrl Https (T.unpack apiUrl) 443 ""
  let makeClientRequest bu req = do
        r <- defaultMakeClientRequest bu req
        pure $ r { requestHeaders = (requestHeaders r) <> [("User-Agent", "servant.client")]}
  let env = env' { makeClientRequest }
  pure $ addLoggingToClientEnv env


fetchConcepts :: Maybe Page
              -> Maybe PerPage
              -> Maybe Cursor
              -> Maybe Filter
              -> Maybe Search
              -> IO (Either ClientError (ListOf Concept))
fetchConcepts mPage mPerPage mCursor mFilter mSearch = do
  env <- defaultClientEnv
  runClientM (concepts mPage mPerPage mCursor mFilter mSearch) env

fetchWorks :: Maybe Page
           -> Maybe PerPage
           -> Maybe Cursor
           -> Maybe Filter
              -> Maybe Search
           -> IO (Either ClientError (ListOf Work))
fetchWorks mPage mPerPage mCursor mFilter mSearch = do
  env <- defaultClientEnv
  runClientM (works mPage mPerPage mCursor mFilter mSearch) env

fetchWorksC :: Maybe Cursor
            -> Maybe Filter
            -> Maybe Search
            -> IO (Either ClientError (Maybe Integer, ConduitT () Work IO ()))
fetchWorksC Nothing mFilter mSearch = do
  fetchWorksC (Just "*") mFilter mSearch
fetchWorksC mCursor mFilter mSearch = do
  env <- defaultClientEnv
  -- NOTE: per-page max is 200
  eRes <- runClientM (works (Just 1) (Just 1) Nothing mFilter mSearch) env
  case eRes of
    Left err -> pure $ Left err
    Right ListOf { meta = Meta { count }} -> do
      pure $ Right ( Just $ fromIntegral count
                   , producer env mCursor )
  where
    producer :: ClientEnv -> Maybe Cursor -> ConduitT () Work IO ()
    producer env mCursor' = do
      let batchSize = 200
      eRes <- liftIO $ runClientM (works Nothing (Just batchSize) mCursor' mFilter mSearch) env
      -- liftIO $ putText $ "Conduit fetching page with cursor " <> show mCursor'
      case eRes of
        Left err -> fail $ "error: " <> show err
        Right ListOf { results, meta = _meta@(Meta { next_cursor }) } -> do
          -- liftIO $ putText $ "Meta: " <> show meta
          --liftIO $ putText $ "[fetchWorksC] Results length: " <> (show $ length results)

          yieldMany results

          if length results < batchSize then
            pure ()
          else do
            producer env next_cursor
